#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

typedef struct big_int_t {
	int n_bits;
	int n_limbs;
	uint64_t *limb;
} big_int;

#define LIMB_MAX	((uint64_t)-1)		// all 1 bits

void big_int_init(big_int *i, int n_bits)
{
	i->n_bits = n_bits;
	i->n_limbs = (n_bits + 64 - 1) / 64;

	i->limb = malloc(i->n_limbs * sizeof(uint64_t));

	if (i->limb == NULL) {
		perror("malloc()");
		exit(1);
	}

	for (int l = 0; l < i->n_limbs; l++) {
		i->limb[l] = 0;
	}
}

void big_int_clear(big_int *i)
{
	free(i->limb);
}

/* big_int_increment() returns:
 *        0 in case of success
 *  nonzero in case of overflow
 */
int big_int_increment(big_int *i)
{
	uint64_t limb_last_max = (LIMB_MAX >> ((i->n_limbs * 64) - i->n_bits));

	for (int l = 0; l < i->n_limbs; l++) {
		if ((l == i->n_limbs - 1) && (i->limb[l] == limb_last_max))
			return -1;

		if (i->limb[l] == LIMB_MAX) {
			assert(l != i->n_limbs - 1);
			i->limb[l] = 0;
			continue;
		}

		i->limb[l] = i->limb[l] + 1;
		return 0;
	}

	assert(0);
	return -1;
}

void big_int_print(big_int *i)
{
	for (int b = i->n_bits - 1; b >= 0; b--) {
		int l = b / 64;
		int lb = b % 64;

		uint64_t mask = ((uint64_t)1) << lb;

		int value;

		if ((i->limb[l] & mask) != 0) {
			value = 1;
		} else {
			value = 0;
		}

		printf("%d", value);
	}
}

int big_int_popcount(big_int *i)
{
	int count = 0;

	for (int l = 0; l < i->n_limbs; l++) {
		count += __builtin_popcount(i->limb[l]);
	}

	return count;
}

int main(int argc, char **argv)
{
	if (argc != 2) {
		printf("Usage: ex1c {number of bits}\n");
		return 1;
	}

	int n = strtol(argv[1], NULL, 0);

	big_int i;

	big_int_init(&i, n);

	while (1) {
		big_int_print(&i);

		printf(": %d\n", big_int_popcount(&i));

		if (big_int_increment(&i))
			break;
	}

	big_int_clear(&i);

	return 0;
}
